#!/usr/bin/env python
#
# piClust
#
# Created by Inuk Jung on 2020-03-14.
# Copyright (c) 2020 Inuk Jung. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division

import sys
# import commands
import math
import decimal

# program error codes
ERROR_CODE=0 	# failed loading expression file
MAPQ=0			# MAPQ filter value (>n)
DEBUG=0			# Debug flag (ON: 1, OFF: 0)

# debug print
def dprint(var_tokens):
	if DEBUG:
		for var in var_tokens: print(var, end=' ')
		print("")

# FLAG masks (SAM format)
FLAG_MULTISEG = 0x1         # template having multiple segments in sequencing
FLAG_PROPALIGN = 0x2        # each segment properly aligned according to the aligner
FLAG_UNMAPPED = 0x4         # segment unmapped
FLAG_NEXTUNMAPPED = 0x8     # next segment in the template unmapped
FLAG_REVCOMP = 0x10         # SEQ being reverse complemented
FLAG_NEXTREVCOMP = 0x20     # SEQ of the next segment in the template being reversed
FLAG_FIRSTSEG = 0x40        # the first segment in the template
FLAG_LASTSEG = 0x80         # the last segment in the template
FLAG_SECONDALIGN = 0x100    # secondary alignment
FLAG_NOPASSQUAL = 0x200     # not passing quality controls
FLAG_PCROPTDUP = 0x400      # PCR or optical duplicate


# Transcript Class
#chrid, tid, tflag, tpos, tqual, tlen, tseq, tmis, status
class Transcript(object):
	def __init__(self, transcript_line):#, tmis=None):
		token = transcript_line.split()
		self.chrid=token[2]			# (str) current processing chromosome
		self.tid=token[0]			# (str) current processing transcript id
		self.tflag=int(token[1])	# (int) transcript bitwise flag (SAM formatted)
		self.tpos=int(token[3])		# (int) transcript mapping pos
#		self.tqual=int(token[4])	# (int) transcript MAPQ value
		self.tlen=len(token[9])		# (int) transcript length
		self.tseq=token[9]			# (str) transcript sequence
#		self.tmis=tmis		# number of mismatch in transcript
		self.status='x'			# status of transcript. Either visited, core or a border transcript (v,c,b)
		self.tstrand=""				# (str) strand of transcript

	def get_t(self, clid):
		return "%s\t%s\t%s\t%d\t%d\t%d\t%s\t%s"%(clid, self.chrid, self.tid, self.tflag, self.tpos, self.tlen, self.tseq, self.tstrand)
	
	def __repr__(self):
		return "%s\t%s\t%d\t%d\t%d\t%s\t%s"%(self.chrid, self.tid, self.tflag, self.tpos, self.tlen, self.tseq, self.tstrand)

# Cluster Class
#chrid, clid, asense_c, sense_c, clscore, cllist[], size(), print
class Cluster(object):
	def __init__(self,chr,clid):
		self.chrid=chr		# clustering chromosome
		self.pos_s=0		# cluster start position
		self.pos_e=0		# cluster end position
		self.clid=clid		# cluster id
		self.asense_c=0		# number of anti-sense strands in cluster
		self.sense_c=0		# number of sense strands in cluster
		self.ntscore=0		# cluster nt bias score
		self.strandscore=0	# cluster's strand score
		self.densityscore=0	# cluster's density score (i.e, log(cldensity/thinnest_cluster_density))
		self.clscore=0		# score of cluster
		self.strand=""		# strand of cluster
		self.cllist=[]		# list of transcripts included in cluster
		self.cldensity=0	# density of cluster
		self.validate=""	# indicates acceptance or rejection of cluster after validation
		self.sense_c=0		# number of sense transcripts
		self.asense_c=0		# number of antisense transcripts

	def add(self, transcript):
		self.cllist.append(transcript)

	def size(self):
		return len(self.cllist)

	def print_cluster(self):
		print("@chr_loc: %s:%d-%d cluster_id: %d cluster_size: %d cluster_score: %f nt: %f density: %f strand: %s"%(self.chrid, self.pos_s, self.pos_e, self.clid, self.size(), self.clscore, self.ntscore, self.cldensity, self.strand))
		# printing list of clustered transcripts
		for t in self.cllist:
			print(t.get_t(self.clid))

	def tofile(self, file):
		file.write("@chr_loc: %s:%d-%d cluster_id: %d cluster_size: %d cluster_score %f nt: %f density: %f strand: %s\n"%(self.chrid, self.pos_s, self.pos_e, self.clid, self.size(), self.clscore, self.ntscore, self.cldensity, self.strand))
		# printing list of clustered transcripts
		for t in self.cllist:
			file.write(t.get_t(self.clid)+"\n")

	def tofasta(self, file):
		for t in self.cllist:
			file.write(">%s\t%s:Cluster_%s\t%s\n%s\n"%(t.tid, self.chrid, self.clid, t.tstrand, t.tseq))

	def __repr__(self):
		# printing brief cluster summary
		print("@chr_loc: %s:%d-%d cluster_id: %d cluster_size: %d cluster_score: %f nt: %f density: %f strand: %s"%(self.chrid, self.pos_s, self.pos_e, self.clid, self.size(), self.clscore, self.ntscore, self.cldensity, self.strand))

def score_cluster(cl, pbio, eps, minpts):
	
	# Transcript class variables: [chrid, tid, tflag, tpos, tqual, tlen, tseq, tmis, status]
	# Cluster class variables: [chrid, clid, asense_c, sense_c, clscore, cldens, cllist[], size(), __repr__]

	# scoring parameters
	entropy_cut=0.35 # entropy cut score for strand bias (e_max=1)

	x_sense_c=0
	x_asense_c=0
	for t in cl.cllist:
		if(t.tflag&FLAG_REVCOMP):	
			x_asense_c+=1
			t.tstrand="-"
		else: 
			x_sense_c+=1
			t.tstrand="+"
	# set strand counts of cluster
	cl.sense_c=x_sense_c
	cl.asense_c=x_asense_c

	x_diverge_count=0 # diverge counter.

	# calculate entropy of strand directionality
	# first test for single strand bias (+/-)
	# check for zero counts
	x_strand="mixed"
	if(x_asense_c==0):
		x_entropy=0
		x_entropy_sense=0
		x_entropy_asense=1
		x_strand="+"
	elif(x_sense_c==0):
		x_entropy=0
		x_entropy_asense=0
		x_entropy_sense=1
		x_strand="-"
	else:
		x_entropy_asense = (x_asense_c/(x_asense_c+x_sense_c))*math.log(x_asense_c/(x_asense_c+x_sense_c),2)*-1
		x_entropy_sense = (x_sense_c/(x_asense_c+x_sense_c))*math.log(x_sense_c/(x_asense_c+x_sense_c),2)*-1
		x_entropy = x_entropy_asense+x_entropy_sense

	# check if single strand bias present
	if (x_entropy < entropy_cut):
		if(x_entropy_asense < x_entropy_sense): x_strand = "-"
		else: x_strand = "+"
	else:
		# if cluster is not single strand biased
		# perform a sliding window divergent strand bias testing
		window_size = int(cl.size()/10)	# the size of the sliding window
		min_window_size=5	# minum window size is 5
		if (window_size < 6):
			window_size = min_window_size
		window_spos = 0
		window_epos = window_spos + window_size

		x_diverge="mixed"
		x_strand="mixed"
		while window_epos < cl.size():
			x_sense_c=0
			x_asense_c=0

			for idx in range(window_size):
				if(cl.cllist[window_spos+idx].tflag&FLAG_REVCOMP): x_asense_c+=1
				else: x_sense_c+=1

			# score entropy for this window
			# check for zero counts
			if(x_asense_c==0):
				x_entropy=0
				x_entropy_sense=0
				x_entropy_asense=12
				if(x_diverge=="-d"):
					x_strand="diverge"
					x_diverge_count+=1
				x_diverge="+d"
				window_spos = window_epos
				window_epos = window_spos + window_size
			elif(x_sense_c==0):
				x_entropy=0
				x_entropy_asense=0
				x_entropy_sense=1
				if(x_diverge=="+d"):
					x_strand="diverge"
					x_diverge_count+=1
				x_diverge="-d"
				window_spos = window_epos
				window_epos = window_spos + window_size
			else:
				x_entropy_asense = (x_asense_c/(x_asense_c+x_sense_c))*math.log(x_asense_c/(x_asense_c+x_sense_c),2)*-1
				x_entropy_sense = (x_sense_c/(x_asense_c+x_sense_c))*math.log(x_sense_c/(x_asense_c+x_sense_c),2)*-1
				x_entropy = x_entropy_asense+x_entropy_sense

				if (x_entropy < entropy_cut):
					if(x_entropy_asense < x_entropy_sense): 
						if(x_diverge=="+d"):
							x_strand="diverge"
						x_diverge = "-d"
					else: 
						if(x_diverge=="-d"):
							x_strand="diverge"
						x_diverge = "+d"

					x_diverge_count+=1

					window_spos = window_epos
					window_epos = window_spos + window_size
				else:
				#	print x_entropy, x_strand, x_diverge, "window", window_spos, x_entropy_asense, x_entropy_sense, cl.clid
					window_spos +=1
					window_epos = window_spos + window_size
					#window_spos = window_epos
					#window_epos = window_spos + window_size

			#print x_strand, x_entropy, window_size, cl.size(), x_entropy_asense, x_entropy_sense, x_sense_c, x_asense_c


	# assign strand to cluster	
	if(x_diverge_count>2): cl.strand="mixed"
	else: cl.strand = x_strand

	#########
	# SCORE 1: score for strand directionality
	# strandscore 2.5~5 (5 if strand bias, 2.5 if mixed)
	strand_score_dict={'diverge':5, '+':5, '-':5, 'mixed':2.5}
	cl.strandscore=strand_score_dict[cl.strand]


	#########
	# SCORE 2: score for 1T or pingpong nucleotide bias
	# ntscore 0~5
	# nt bias analysis (counting 1T and 10A sequences)
	nt_count_1st={'A':0, 'T':0, 'G':0, 'C':0}
	nt_count_10th={'A':0, 'T':0, 'G':0, 'C':0}
	nt_count_1T10A=0
	
	for t in cl.cllist:
		if t.tseq[0] in nt_count_1st and t.tseq[9] in nt_count_10th:
			nt_count_1st[t.tseq[0]] +=1
			nt_count_10th[t.tseq[0]] +=1
			if t.tseq[0]=='T' and t.tseq[9]=='A':
				nt_count_1T10A+=1
#		else:		# non ATGC nulceotid, N/X or any other ambiguous characters
#			print t.tseq
#	print cl.clid, len(cl.cllist), nt_count_1st['A'],nt_count_1st['T'],nt_count_1st['G'],nt_count_1st['C']
#	print cl.clid, len(cl.cllist), nt_count_10th['A'],nt_count_10th['T'],nt_count_10th['G'],nt_count_10th['C']

	if pbio=="pingpong":
		# given score according 1T and 10A only
		cl.ntscore = nt_count_1T10A/len(cl.cllist) * 5
	else:
		# given score to 1T (10A inclusive)
		cl.ntscore = nt_count_1st['T']/len(cl.cllist) * 5
	
	#########
	# SCORE 3: score for density and size of cluster
#	print minpts, eps, cl.cldensity
#	if cl.cldensity==0:
#		for t in cl.cllist:
#			print cl.clid, t

	thinnest_cluster = minpts/eps*1000	# 'n' piRNAs/KB, the density of thinnest cluster
	#cl.densityscore = math.log(cl.cldensity/thinnest_cluster)
	cl.densityscore = cl.cldensity/thinnest_cluster

	
	#########
	# Total score: SCORE 1 + 2 + 3
	cl.clscore = cl.strandscore + cl.ntscore + cl.densityscore		
	#print cl.clscore, cl.strandscore, cl.ntscore, cl.densityscore
	

	return cl.clscore	


def cluster(chr, eps, minpts, c_score, clf, uclf, clfa, uclfa, cur_clid, pbio):

	# initialize transcript cluster
	clid=cur_clid
	picluster=Cluster(chr.chrid, clid)
	
	dist=0
	coreidx = 0
	cand_core = chr.cllist[coreidx]
	picluster.add(chr.cllist[coreidx])
	
	
	rej_clid=1	# index for rejected clusters
	i=1 # index for accepted clusters
	while (i < chr.size()):

		dist = chr.cllist[i].tpos - cand_core.tpos
		#print "dist[%d]: %d, (%d - %d --> %d - %d)"%(i,dist, chr.cllist[i].tpos, cand_core.tpos, i, coreidx)
		dprint(["start: ", coreidx, i])
		dprint([">>", chr.cllist[coreidx]])
		dprint([">", chr.cllist[i]])

		if(dist<=eps):
			dprint(["eps good"])
			if(chr.cllist[i].status!='v'):
				chr.cllist[i].status='v'	# update transcript status as visited
				picluster.add(chr.cllist[i])
			i+=1 # proceed iteration
		elif((i-coreidx)>minpts): # and chr.cllist[i].status!='b'):
			dprint(["eps not good, minpts good"])
			dprint([coreidx, i, minpts])
			dprint([chr.cllist[i]])
			dprint([chr.cllist[coreidx]])

			# update candidate core transcript
			if(picluster.size()>minpts):
				coreidx=i-minpts
				cand_core = chr.cllist[coreidx]
			else:
				coreidx+=1
				cand_core = chr.cllist[coreidx]

			#print "expanding with index %d, current index %d, total %d"%(coreidx, i, chr.size())
			#chr.cllist[i-1].status='b'	# update transcript status as border

			#DEBUG
			i = coreidx+1	# updating 'i' index to coreidx for next cluster expansion index

		else:
			dprint(["eps not good"])
			# verify if cluster has minpts
			if(picluster.size()>minpts):
				dprint(["eps not good, clusters minpts good"])

				#print "chr: %s cluster_id: %d cluster_size: %d cluster_score %f nt:%f, e:%f"%(picluster.chrid, picluster.clid, picluster.size(), picluster.clscore, picluster.ntscore, picluster.entropy)

				picluster.pos_s=picluster.cllist[0].tpos # set cluster start position
				picluster.pos_e=picluster.cllist[len(picluster.cllist)-1].tpos + picluster.cllist[len(picluster.cllist)-1].tlen	# set cluster end position

				# calculate cluster density
				cllength = (picluster.pos_e - picluster.pos_s)/1000		# units in KB
				picluster.cldensity = len(picluster.cllist)/cllength	# piRNA/KB
		
				# score candidate cluster
				score = score_cluster(picluster,pbio,eps,minpts)

				if score >= c_score:
					picluster.validate="accept"
					picluster.tofile(clf) # writing accepted clusters to file
					picluster.tofasta(clfa) # writing accepted clusters to file in fasta format
					clid+=1 # increment cluster id

					picluster=Cluster(chr.chrid, clid) # initiate new cluster
					picluster.add(chr.cllist[i]) # add current node to new cluster
				else:
					# cluster does not pass minimum score
					picluster.clid = rej_clid
					rej_clid+=1
					picluster.validate="reject"

					#DEBUG
					dprint(["REJECT-1"])

					picluster.pos_s=picluster.cllist[0].tpos # set cluster start position
					picluster.pos_e=picluster.cllist[len(picluster.cllist)-1].tpos + picluster.cllist[len(picluster.cllist)-1].tlen	# set cluster end position

					picluster.tofile(uclf) # writing rejected clusters to file
					picluster.tofasta(uclfa) # writing rejected clusters to file in fasta format
					picluster=Cluster(chr.chrid, clid)
					picluster.add(chr.cllist[i]) # add rejected node to new cluster

			else:
				# cluster does not pass minpts

				#DEBUG
				dprint(["REJECT-2"])

				picluster.clid = rej_clid
				rej_clid+=1
				picluster.validate="reject"
				picluster.pos_s=picluster.cllist[0].tpos # set cluster start position
				picluster.pos_e=picluster.cllist[len(picluster.cllist)-1].tpos + picluster.cllist[len(picluster.cllist)-1].tlen	# set cluster end position

				picluster.tofile(uclf) # writing rejected clusters to file
				picluster.tofasta(uclfa) # writing rejected clusters to file in fasta format
				picluster=Cluster(chr.chrid, clid)
				picluster.add(chr.cllist[i]) # DEBUG
				

			# update core transcript
			coreidx=i
			cand_core = chr.cllist[coreidx]

			# take next alignment line
			i += 1

			
	# verify last cluster
	if(picluster.size()>=minpts):

		# accepted clusters
		#print "chr: %s cluster_id: %d cluster_size: %d cluster_score %f nt:%f, e:%f"%(picluster.chrid, picluster.clid, picluster.size(), picluster.clscore, picluster.ntscore, picluster.entropy)
		picluster.pos_s=picluster.cllist[0].tpos # set cluster start position
		picluster.pos_e=picluster.cllist[len(picluster.cllist)-1].tpos + picluster.cllist[len(picluster.cllist)-1].tlen	# set cluster end position

		# calculate cluster density
		cllength = (picluster.pos_e - picluster.pos_s)/1000		# units in KB
		picluster.cldensity = len(picluster.cllist)/cllength	# piRNA/KB

		score = score_cluster(picluster,pbio,eps,minpts)
		if score >= c_score:
			picluster.validate="accept"
			picluster.tofile(clf) # writing accepted clusters to file
			picluster.tofasta(clfa) # writing accepted clusters to file in fasta format
			clid+=1 # increment cluster id
		else:
			#print "REJECT 3"
			picluster.validate="reject"
			# rejected clusters
			picluster.tofile(uclf) # writing rejected clusters to file
			picluster.tofasta(uclfa) # writing rejected clusters to file in fasta format

	return clid

def main():
	
	# setting cluster variable
	eps=int(sys.argv[1]) 			# epsilon value
	minpts=int(sys.argv[2])			# minpts value (minimum number of points for core point condition)
	core_dist=0						# distance from core to a neighbor point
	clid=0							# cluster id
	chr_transcripts=Cluster("init", clid)			# chromosome object (includes transcripts per chromosome)
	#score_params=2					# default 2 scoring parameters (+1 optional if expression file provided)
	tmp_core_buffer=[]

	# setting piRNA biogenesis model
	#piRNA_biogen="pingpong"	# could be set to either 'pingpong' or 'na'
	piRNA_biogen="unknown"	# could be set to either 'pingpong' or 'na'

	# setting in/output variables	
	line_counter=0					# input file's line counter
	infile = open(sys.argv[4], 'r') # intput file

	#out_dir = commands.getoutput('dirname ' + sys.argv[4]) + "/cluster_out_" + sys.argv[5]
	out_dir = sys.argv[5]

	cluster_outfile = open(out_dir+"/cluster_out.dat", 'w')			# cluster output file
	cluster_outfasta = open(out_dir+"/cluster_out.fasta", 'w')         # cluster output file in fasta format
	uncluster_outfile = open(out_dir+"/unclustered_out.dat", 'w')	# unclustered output file
	uncluster_outfasta = open(out_dir+"/unclustered_out.fasta", 'w')	# unclustered output file


	# setting report variables
	as_strand_c=0	# anti-sense counter
	s_strand_c=0	# sense counter
	grand_total_clusters=0			# grand total number of clusters
	grand_total_clustered_reads=0	# grand total number of clustered sequences (among all reads)
	grand_total_reads=0				# grand total number of aligned reads
	cut_score = float(sys.argv[3]) 	# cluster cut score

	
	# DEBUG
#	cut_score=2


	line = infile.readline()
	while True:
		if not line:
			break

		
		if (line[0] != "@"): 				# skipping bowtie headers
			if int(line.split()[4])>MAPQ:	# filter read by mapping quality
				tmp_core = Transcript(line.strip()) # loading transcript

				if(chr_transcripts.chrid != tmp_core.chrid):# initiating a new cluster
					
					if(chr_transcripts.chrid != "init"): 	# do summary of current cluster

						# cluster transcripts of chromosome chr_transcripts.chr
						clid = cluster(chr_transcripts, eps, minpts, cut_score, cluster_outfile, uncluster_outfile,cluster_outfasta, uncluster_outfasta,clid,piRNA_biogen)
		
						# increment clid
						#clid=cur_clid

						# initialize next chromosome cluster
						chr_transcripts = Cluster(tmp_core.chrid, clid)

					else:	# iniating first cluster
						chr_transcripts.chrid = tmp_core.chrid
						chr_transcripts.clid = clid
					
			
					# validate tmp_core for core point condition
					chr_transcripts.add(tmp_core)
			
				else:
					# proceed with clustering within chromosome cur_chr
					# validate tmp_core for core point condition
					chr_transcripts.add(tmp_core)
				

		line = infile.readline()

	# perform clustering for the last chromosome in list
	cluster(chr_transcripts, eps, minpts, cut_score, cluster_outfile, uncluster_outfile,cluster_outfasta, uncluster_outfasta, clid, piRNA_biogen)
		

	# Closing files	
	infile.close()
	cluster_outfile.close()
	cluster_outfasta.close()
	uncluster_outfile.close()
	uncluster_outfasta.close()
	
if __name__ == "__main__":
	main()



